#!/bin/bash
########################################################################
# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# @(#) Micorosoft SQL Server on Linux Attribute Definition File
# @(#) Product Name                :  HPE Serviceguard extension for Microsoft SQL Server. SGeMSSQL
# @(#) Product Version             :  A.12.20.00
# @(#) Patch Name                  :  
#
# *** Note: This file MUST NOT be edited. ***
#
# These attributes are used for Microsoft SQL Server on Linux package and
# should not be manually edited. Any changes made to it will
# be overwritten when you upgrade to the next release of Serviceguard Extension
# for Microsoft SQL Server
#
#-----------------------------------------------------------------------
# Description:  Microsoft SQL Server on Linux Modular package configuration script.
#
# Input : The script accepts two parameters: op_code and filename.
#
#    op_code :  The first argument is the operation to be performed whose
#               valid values are: start, stop, ppas_monitor, validate,
#               reconfigure_start and reconfigure_stop.
#               "start":     Calls mssql scripts with 'start' option.
#                            If error occurs, execute each previous module
#                            script with stop parameter in reverse order.
#               "stop" :     Calls each of the module scripts with 'stop'
#                            option.
#               "monitor": Calls mssql scripts with 'monitor' option.
#
#               "validate":  Calls each of the module scripts with 'validate'
#                            option.
#               "reconfigure_start": Calls each of the module scripts with
#                            'reconfigure_start' option
#               "reconfigure_stop": Calls each of the module scripts with
#                            'reconfigure_stop' option
#               "port_monitor": Calls mssql script with 'port_monitor'
#                               option, to monitor the mssql Server port.
#
# Output : Errors/Log messages.
#          In case of "validate" the messages are printed to stdout.
#          For "start" and "stop" cases the messages are directed to
#          $SG_SCRIPT_LOG_FILE. If the SCRIPT_LOG_FILE is not specified
#          the messages go to ${SGRUN}/log/${SG_PACKAGE_NAME}.log.
#
# Exit:     0  When the execution of all the scripts is successful.
#           1  When there is an error condition from the script and the
#              package should not be restarted as a result of the error.
#           2  When there is an error condition from the script, but the
#              package can be restarted on an alternate node.
#
#########################################################################

#
# Source in Serviceguard environment variables
#
. /etc/cmcluster.conf

#
# Source in the utility functions used by the module scripts
#
if [[ -z $SG_UTILS ]]
then
    SG_UTILS=$SGCONF/scripts/mscripts/utils.sh
fi
if [[ -f ${SG_UTILS} ]]
then
    . ${SG_UTILS}
    if [[ $? -ne 0 ]]
    then
        echo "ERROR: Unable to source package utility functions file: ${SG_UTILS}"
        exit 1
    fi
else
    echo "ERROR: Unable to find package utility functions file: ${SG_UTILS}"
    exit 1
fi

SGEMSSQL_UTILS=${SGCONF}/scripts/sgemssql/mssql_utils.sh
export SGEMSSQL_UTILS
if [[ -f ${SGEMSSQL_UTILS} ]]; then
    . ${SGEMSSQL_UTILS}
    if (( $? != 0 ))
    then
        echo "ERROR: Unable to source SGeMSSQL utility functions file: ${SGEMSSQL_UTILS}"
        exit 1
    fi
else
    echo "ERROR: Unable to find SGeMSSQL utility functions file: ${SGEMSSQL_UTILS}"
    exit 1
fi

#
# PATH for Linux
#
PATH=$SGSBIN:/usr/bin:/usr/sbin:/etc:/bin:/sbin:/usr/lbin
export PATH

#
# Get the environment for this package through utility function
# sg_source_pkg_env().
#
sg_source_pkg_env $*

#
# Get parameters
#
if [[ $# == 0 ]]
then
   sg_log 0 "ERROR: No parameter is passed to the command."
   sg_exit 1 $*
fi

#
# Export the global variable op_code
#
typeset -x op_code="${1}"

typeset -i exit_val=0
typeset -i mssql_srv_restart=0

KILL=/bin/kill
PS=/bin/ps
AWK=/bin/awk
SLEEP=/usr/bin/sleep

###############################################################################
# Function: mssql_server_pkg_validate
#
# Validate the package attributes and also check if Microsoft SQL server is
# configured.
# Return 0 on success otherwise return 1
###############################################################################
function mssql_server_pkg_validate
{
    typeset -i _ret=0

    cache_clview_sconfig

    if [[ -z $SG_PRIV_CLUSTER_VERIFICATION  || $SG_COMMAND = "cmapplyconf" ]]; then
        # Check if package is applied
        $cmviewcl -f line -s config -p $SG_PACKAGE > /dev/null 2>&1
        if [[ $? -eq 0 ]]; then
            # Check if package is running and check if this modules is not newly getting
            # added during online configuration, then only call function verify_unsupported_online_modification
            pkg_stat=`$cmviewcl -f line | grep "$SG_PACKAGE|status" | cut -d "=" -f 2 2>&1`
            if [[ "$pkg_stat" = "up" && ${PREV_SG_MODULE_NAME[@]} = *"sgemssql/mssqlserver"* ]]; then
                verify_unsupported_online_modification aoai
                if [[ $? -eq 1 ]]; then
                    return 1
                fi
            fi
         fi
    fi
    # Validate if Microsoft SQL Server is installed
    validate_mssql_server_configuration
    if [[ $? -ne 0 ]]
    then
        _ret=1
    fi

    # Validate if the login credentials are valid.
    if pgrep -x "sqlservr" > /dev/null
    then
        validate_login_credentials_for_package $MSSQL_MSSQL_ADMIN $MSSQL_MSSQL_PASSWORD\
                                               $LOCAL_HOST_NAME
        if [[ $? -ne 0 ]]
        then
            _ret=1
        fi
    else
        sg_log 0 "WARNING: MSSQL credential validation skipped, since MSSQL Server instance in not"\
                 "running on the node \"$LOCAL_HOST_NAME\"."
    fi

    # Validate if Microsoft SQL Server package is already exists
    does_mssql_server_package_exist
    if [[ $? -ne 0 ]]
    then
        _ret=1
    fi

    #validate the MSSQL port
    mssql_port=`get_mssql_tcp_port`
    if [[ "$mssql_port" != "$MSSQL_MSSQL_PORT" ]]; then
       sg_log 0 "ERROR: Port specified in the package configuration file does not match"\
                "the port configured for MSSQL server."
       _ret=1
    fi

    exit_val=$_ret
}

###############################################################################
# Function: mssql_start_cmd
#
# Start Microsoft SQL Server on Linux Instance if it is not running already.
# Return 0 on success otherwise return 1
###############################################################################
function mssql_start_cmd
{
    # See if the Microsoft SQL Server exist.
    details=`rpm -qi mssql-server`
    status=$?
    if [[ $status -ne 0 ]]; then
        sg_log 0 "ERROR: Microsoft SQL Server on Linux is not Installed"
        return 1
    fi

    # Check if Microsoft SQL Server on linux is already running.
    PID=`systemctl status mssql-server | grep "Main PID:" | grep -v "code=exited" | awk '{print $3}'`
    status=$?
    if [[ $status -ne 0 ]]; then
        # Is already running
        sg_log 0 "Microsoft SQL sever on Linux is already running with PID=$PID"
        return 0
    fi

    # Microsoft SQL Server on Linux is not running
    sg_log 0 "Starting Microsoft SQL Server on Linux:"

    systemctl start mssql-server
    status=$?

    if [[ $status -eq 0 ]]; then
        sg_log 0 "Microsoft SQL Server on Linux Started Successfully"
    else
        sg_log 0 "ERROR: Microsoft SQL Server on Linux failed to start check the log file for details."
        return 1
    fi

    return 0
}

###############################################################################
# Function: mssql_stop_cmd
#
# Stop Microsoft SQL Server on Linux Instance if it is not running already.
# Return 0 on success otherwise return 1
###############################################################################
function mssql_stop_cmd
{
    #Check if Microsoft SQL Server on Linux is already stopped
    PID=`systemctl status mssql-server | grep "Main PID:" | grep -v "code=exited" | awk '{print $3}'`

    if [[ -z $PID ]] ; then
        sg_log 0 "Microsoft SQL Server on Linux is not running"
        return 0
    fi

    sg_log 0 "Stopping Microsoft SQL Server On Linux:"
    systemctl stop mssql-server

    # See if it is still running.
    PID=`systemctl status mssql-server | grep "Main PID:" | grep -v "code=exited" | awk '{print $3}'`

    if [[ $PID ]] ; then
        sg_log 0 "Graceful shutdown of Microsoft SQL Server on Linux failed. Killing the process."
        $PS -aef | sqlservr | $AWK '{print $2}' | xargs $KILL -9
    fi
    sg_log 0 "Microsoft SQL Server on Linux Stopped"

    return 0
}

#######################################################################
# Function: validate_reconfiguration_restart
#
# Validate if the online reconfiguration requires the monitoring script
# to be restarted.
#######################################################################
function validate_reconfiguration_restart
{

    #Check if all the services needs to be restarted
    if [[ "${MSSQL_MONITOR_INTERVAL}" != "${PREV_MSSQL_MONITOR_INTERVAL}" ]] ||
       [[ ${SG_EMAIL_ID[*]} !=  ${PREV_SG_EMAIL_ID[*]} ]] ||
       [[ ${SG_SCRIPT_LOG_FILE} != ${PREV_SG_SCRIPT_LOG_FILE} ]]; then
           mssql_srv_restart=1
    fi
    
    if [[ $mssql_srv_restart -eq 1 ]]; then
        return 0
    else
        return 1
    fi
}

###############################################################################
# Function: reconfigure_stop_function
#
# Validate the whether reconfigure is valid. If so, halt the service script
# if not, do not allow restart of  service script.
# Return 0 on success otherwise prints appropriate error and return 1
###############################################################################
function reconfigure_stop_function
{
    exit_val=0
    if sg_check_if_reconfig_has_failed
    then
        sg_log 999 "WARNING: Not attempting to stop the MSSQL services for $SG_PACKAGE" \
                   "during online reconfiguration stop."
        return 0
    fi

    validate_reconfiguration_restart
    if [[ $? -eq 0 ]]
    then
        #Restart all services of the aoai if common attribute are changed.
        if [[ $mssql_srv_restart -eq 1 ]]; then
            for (( i=0; i<${#PREV_SG_SERVICE_NAME[@]}; i++ ))
            do
                if ! `sg_check_if_reconfig_has_failed` ; then
                    ${SGSBIN}/cmhaltserv ${PREV_SG_SERVICE_NAME[$i]}
                    if [[ $? -ne 0 ]]; then
                        sg_log 999 "ERROR: Failed to stop the MSSQL toolkit service ${PREV_SG_SERVICE_NAME[$i]}"
                        sg_set_reconfig_failure
                        exit_val=1
                    else
                        sg_log 999 "SUCCESS: Successfully stopped MSSQL toolkit service ${PREV_SG_SERVICE_NAME[$i]}."
                        sg_log 0 "SUCCESS: Successfully stopped MSSQL toolkit service ${PREV_SG_SERVICE_NAME[$i]}."
                    fi
                else
                    sg_log 999 "WARNING: Not attempting to stop the MSSQL toolkit service ${PREV_SG_SERVICE_NAME[$i]}."
                fi 
            done
        fi
    fi
    return $exit_val
}

###############################################################################
# Function: reconfigure_start_function
#
# Validate the whether reconfigure is valid. If so, run the service script
# if not, do not allow restart of service script.
# Return 0 on success otherwise prints appropriate error and return 1
###############################################################################
function reconfigure_start_function
{
    exit_val=0
    if sg_check_if_reconfig_has_failed
    then
        sg_log 999 "WARNING: Not attempting to start the MSSQL toolkit services for $SG_PACKAGE" \
                   "during online reconfiguration start."
        return 0
    fi
    validate_reconfiguration_restart
    if [[ $? -eq 0 ]]
    then
        if [[ $mssql_srv_restart -eq 1 ]]; then
            for (( i=0; i<${#SG_SERVICE_NAME[@]}; i++ ))
            do
                msg="cmrunserv: Unable to start service ${SG_SERVICE_NAME[$i]} (it is already active)"
                if ! `sg_check_if_reconfig_has_failed` ; then
                    out=`${SGSBIN}/cmrunserv ${SG_SERVICE_NAME[$i]} "${SG_SERVICE_CMD[$i]}" 2>&1`
                    ret=$?
                    if [[ $ret -ne 0 ]] && [[ "$out" != "$msg" ]]; then
                        sg_log 999 "ERROR: Failed to start the MSSQL toolkit service ${SG_SERVICE_NAME[$i]}"
                        sg_set_reconfig_failure
                        exit_val=1
                    elif [[ $ret -eq 0 ]]; then
                        sg_log 999 "SUCCESS: Successfully started MSSQL toolkit service ${SG_SERVICE_NAME[$i]}."
                        sg_log 0 "SUCCESS: Successfully started MSSQL toolkit service ${SG_SERVICE_NAME[$i]}."
                    fi
                else    
                    sg_log 999 "WARNING: Not attempting to start MSSQL toolkit service ${SG_SERVICE_NAME[$i]}."   
                fi
            done
        fi            
    fi

    return $exit_val
}

get_user_name_and_password_from_login_credentials_file
if [ $? -ne 0 ]; then
    sg_log 0 "ERROR: MSSQL Credential validation failed on the"\
             "node \"$LOCAL_HOST_NAME\""
    exit 1
fi

case "${op_code}" in
    "start")
        if sg_check_if_reconfig_has_failed ; then
            sg_log 999 "WARNING: Not attempting to start Microsoft SQL Server on Linux module"
            return
        else
            sg_log 0 "Starting Microsoft SQL Server On Linux.."
            mssql_start_cmd
            check_status=$?
            if [[ $check_status -ne 0 ]] ; then
                if sg_check_if_it_is_reconfiguration ; then
                    sg_log 999 "ERROR: Failed to start Microsoft SQL Server on Linux module"
                    sg_set_reconfig_failure
                    exit_val=1
                fi
            else
               if sg_check_if_it_is_reconfiguration ; then
                   sg_log 999 "SUCCESS: Successfully started Microsoft SQL Server on Linux module"
               fi
            fi
        fi
        ;;
    "stop")
        sg_log 0 "Stopping Microsoft SQL Server on Linux"
        mssql_stop_cmd
        ;;
    "monitor")
        sg_log 0 "Starting Monitoring for Microsoft SQL Server on Linux"
        . $SGSBIN/mssql/hamssql.mon
        ;;
    "health")
        sg_log 0 "Starting Health Monitoring for Microsoft SQL Server on Linux"
        . $SGSBIN/mssql/sp_server_diagnostics.sh
        ;;
    "validate")
        sg_log 0 "Validate the Microsoft SQL Server parameters"
        mssql_server_pkg_validate
        ;;
    "port_monitor")
        sg_log 0 "Starting Port Monitoring for Microsoft SQL Server on Linux"
        . $SGSBIN/mssql/mssql_port.mon
        ;;
    "reconfigure_start")
        reconfigure_start_function
        ;;
    "reconfigure_stop")
        reconfigure_stop_function
        ;;
    *)
        sg_log 0 "Usage: ${0} [ start | stop | validate | monitor | port_monitor | \
                         reconfigure_start | reconfigure_stop ]"
        ;;
esac

exit $exit_val
