#!/bin/bash
##################################################################################
# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# @(#) Micorosoft SQL Server on Linux Attribute Definition File
# @(#) Product Name                :  HPE Serviceguard extension for Microsoft SQL Server. SGeMSSQL
# @(#) Product Version             :  A.12.20.00
# @(#) Patch Name                  :  
#
# *** Note: This file MUST NOT be edited. ***
#
# These attributes are used for Microsoft SQL Server on Linux package and
# should not be manually edited. Any changes made to it will
# be overwritten when you upgrade to the next release of Serviceguard Extension
# for Microsoft SQL Server
#
#-----------------------------------------------------------------------
# Description: Microsoft SQL Server on Linux Modular package configuration script.
#
# Input : The script accepts two parameters: op_code and filename.
#
#    op_code :  The first argument is the operation to be performed whose
#               valid values are: start, stop, monitor, health, validate,
#               instance_monitor, reconfigure_start and reconfigure_stop.
#               "start"    : Calls mssql scripts with 'start' option.
#                            If error occurs, execute each previous module
#                            script with stop parameter in reverse order.
#
#               "stop"     : Calls each of the module scripts with 'stop'
#                            option.
#
#               "monitor"  : Calls mssql scripts with 'monitor' option.
#                            For monitoring mssql Server Process.
#
#               "health"   : Calls mssql tookkit script with 'health' option.
#                            to obtain the sp_server_health_diagnostics.
#
#               "validate" :  Calls each of the module scripts with 'validate'
#                            option.
#
#               "reconfigure_start": Calls each of the module scripts with
#                            'reconfigure_start' option
#
#               "reconfigure_stop": Calls each of the module scripts with
#                            'reconfigure_stop' option
#
#               "replica_monitor": Calls mssql script to monitor local node
#                             role with option "replica_monitor" and the
#                             generic resource which is used by it
#
# Output : Errors/Log messages.
#          In case of "validate" the messages are printed to stdout.
#          For "start" and "stop" cases the messages are directed to
#          $SG_SCRIPT_LOG_FILE. If the SCRIPT_LOG_FILE is not specified
#          the messages go to ${SGRUN}/log/${SG_PACKAGE_NAME}.log.
#
# Exit:     0  When the execution of all the scripts is successful.
#           1  When there is an error condition from the script and the
#              package should not be restarted as a result of the error.
#           2  When there is an error condition from the script, but the
#              package can be restarted on an alternate node.
#
##################################################################################

#
# Source in Serviceguard environment variables
#
. /etc/cmcluster.conf

#
# Source in the utility functions used by the module scripts
#
if [[ -z $SG_UTILS ]]
then
    SG_UTILS=$SGCONF/scripts/mscripts/utils.sh
fi
if [[ -f ${SG_UTILS} ]]
then
    . ${SG_UTILS}
    if [[ $? -ne 0 ]]
    then
        echo "ERROR: Unable to source package utility functions file: ${SG_UTILS}"
        exit 1
    fi
else
    echo "ERROR: Unable to find package utility functions file: ${SG_UTILS}"
    exit 1
fi

SGEMSSQL_UTILS=${SGCONF}/scripts/sgemssql/mssql_utils.sh
export SGEMSSQL_UTILS
if [[ -f ${SGEMSSQL_UTILS} ]]; then
    . ${SGEMSSQL_UTILS}
    if (( $? != 0 ))
    then
        echo "ERROR: Unable to source SGeMSSQL utility functions file: ${SGEMSSQL_UTILS}"
        exit 1
    fi
else
    echo "ERROR: Unable to find SGeMSSQL utility functions file: ${SGEMSSQL_UTILS}"
    exit 1
fi

#
# PATH for Linux
#
PATH=$SGSBIN:/usr/bin:/usr/sbin:/etc:/bin:/sbin:/usr/lbin
export PATH

#
# Get parameters
#
if [[ $# == 0 ]]
then
   sg_log 0 "ERROR: No parameter is passed to the command."
   sg_exit 1 $*
fi

case $1 in
    'validate'|'start'|'stop'|'reconfigure_start'|'reconfigure_stop')
        sg_source_pkg_env $* ;;
    *)
        # Since sg_source_pkg_env expects more than one value, it considers $2 in this case
        # as file and fails, so lets pass just $1 which should work fine.
        sg_source_pkg_env $1 ;;
esac



###############################################################################
# Function: mssql_validate_cmd
#
# Validate the package attributes and also check if Microsoft SQL server is
# configured.
# Return 0 on success otherwise return 1
###############################################################################
function mssql_validate_cmd
{
    typeset -i _ret=0

    cache_clview_sconfig

    # Validate if HADR enabled
    is_aoai_enabled
    if [[ $? -ne 0 ]]
    then
        sg_log 0 "ERROR: HADR is not enabled"
        _ret=1
    fi

    # Validate if the login credentials are valid.
    if pgrep -x "sqlservr" > /dev/null
    then
        validate_login_credentials_for_package $MSSQL_MSSQL_ADMIN $MSSQL_MSSQL_PASSWORD\
                                               $LOCAL_HOST_NAME
        if [[ $? -ne 0 ]]
        then
            _ret=1
        fi
    else
        sg_log 0 "WARNING: MSSQL Credential validation skipped, since MSSQL Server instance in not"\
                 "running on the node \"$LOCAL_HOST_NAME\". "
    fi

    # Validate the if AOAI package is already configured with the provided AG:
    check_if_ag_already_configured_in_a_package
    if [[ $? -ne 0 ]]
    then
        _ret=1
    fi

    # Validate the configured parameters:
    validate_aoai_parameters
    if [[ $? -ne 0 ]]
    then
        _ret=1
    fi

    is_package_up ${SG_PACKAGE_NAME}
    if [[ $? -eq 0 ]]
    then
        # This means the package is running
        # There is only one AG name that can be provided in an AOAI package
        # Therefore we use array index 0.
        if [[ "${MSSQL_AVAILABILITY_GROUP[0]}" != "${PREV_MSSQL_AVAILABILITY_GROUP[0]}" ]]
        then
            sg_log 0 "ERROR: Online modification of Availability Group of" \
                       "Microsoft SQL Server Always ON Availability instance" \
                       "package is not supported."
            _ret=1
        fi

        # Disallow online deletion of a replica
        for prev_node in ${PREV_MSSQL_NODENAME[@]}
        do
            if  ! `echo "${MSSQL_NODENAME[@]}" | grep -w "$prev_node" > /dev/null`
            then
                sg_log 0 "ERROR: Online deletion of a replica $prev_node in" \
                       "Microsoft SQL Server Always ON Availability instance" \
                       "package is not supported."
                _ret=1
            fi
        done

    fi

    exit_val=$_ret
}

###############################################################################
# Function: mssql_start_cmd
#
###############################################################################
function mssql_start_cmd
{
    AG=$1
    # Create sgmssql to store the start_lock_file
    file_dir=$SGRUN/sgemssql
    if [[ ! -d $file_dir ]]; then
        mkdir $file_dir
    fi
    # File Name used for locking during the start of AOAI MNP package
    START_FILE="$SGRUN/sgemssql/."$AG"_start_file.lck"
    if [[ ! -f $START_FILE ]]; then
        touch $START_FILE
    fi
}

###############################################################################
# Function: mssql_stop_cmd
#
###############################################################################
function mssql_stop_cmd
{
    # Place holder
    :
}

#######################################################################
# Function: validate_reconfigure_restart
#
# Validate if the online reconfiguration requires the monitoring script
# to be restarted.
#######################################################################
function validate_reconfigure_restart
{

    if [[ $SG_SCRIPT_LOG_FILE = $PREV_SG_SCRIPT_LOG_FILE ]] &&
       [[ ${SG_EMAIL_ID[*]} =  ${PREV_SG_EMAIL_ID[*]} ]]
    then
        return 0
    fi
    return 1
}

###############################################################################
# Function: reconfigure_stop_function
#
# Validate the whether reconfigure is valid. If so, halt the service script
# if not, do not allow restart of  service script.
# Return 0 on success otherwise prints appropriate error and return 1
###############################################################################
function reconfigure_stop_function
{

    if sg_check_if_reconfig_has_failed
    then
        sg_log 999 "WARNING: Not attempting to stop the Microsoft SQL Server"\
                   "Always ON Availability instance Monitoring service"\
                   "for online package modification."
        return
    fi

    validate_reconfigure_restart
    if [[ $? -ne 0 ]]
    then
        for (( i=0; i<${#PREV_SG_SERVICE_NAME[@]}; i++ ))
        do
            if [[ "`echo ${PREV_SG_SERVICE_CMD[$i]} | $AWK '{print $2}'`" = "replica_monitor" ]]; then
                if ! ${SGSBIN}/cmhaltserv ${PREV_SG_SERVICE_NAME[$i]} ; then
                    sg_log 999 "ERROR: Could not stop the service "\
                               "${PREV_SG_SERVICE_NAME[$i]} for online package"\
                               "modification."
                    sg_set_reconfig_failure
                    exit_val=1
                else
                    sg_log 999 "SUCCESS: Successfully stopped the service"\
                               "${PREV_SG_SERVICE_NAME[$i]} online package"\
                               "modification."
                    if ! sg_check_if_reconfig_has_failed ; then
                         sg_log 0 "SUCCESS: Successfully stopped the service"\
                                 "${PREV_SG_SERVICE_NAME[$i]} for online"\
                                 "package modification."
                    fi
                fi
            fi
        done
    fi

    return $exit_val
}

###############################################################################
# Function: reconfigure_start_function
#
# Validate the whether reconfigure is valid. If so, run the service script
# if not, do not allow restart of service script.
# Return 0 on success otherwise prints appropriate error and return 1
###############################################################################
function reconfigure_start_function
{
    validate_reconfigure_restart
    if [[ $? -ne 0 ]]
    then
        for (( i=0; i<${#SG_SERVICE_NAME[@]}; i++ ))
        do
            if [[ "`echo ${SG_SERVICE_CMD[$i]} | $AWK '{print $2}'`" = "replica_monitor" ]]; then
                msg="cmrunserv: Unable to start service ${SG_SERVICE_CMD[$i]} (it is already active)"
                if ! `sg_check_if_reconfig_has_failed` ; then
                    out=`${SGSBIN}/cmrunserv ${SG_SERVICE_NAME[$i]} "${SG_SERVICE_CMD[$i]}" 2>&1`
                    exit_code=$?
                    if [[ $exit_code -ne 0 ]] && [[ "$msg" != "$out" ]] ;
                    then
                        sg_log 999 "ERROR: Failed to start the service"\
                                   "${SG_SERVICE_NAME[$i]} for online package"\
                                   "modification."
                        sg_set_reconfig_failure
                        exit_val=1
                    elif [[ $exit_code -eq 0 ]] ; then
                        sg_log 999 "SUCCESS: Successfully started the service"\
                                   "${SG_SERVICE_NAME[$i]} for online package"\
                                   "modification."
                         sg_log 0 "SUCCESS: Successfully started service"\
                                  "${SG_SERVICE_NAME[$i]} for online package"\
                                  "modification."
                    fi
                fi
            fi
        done
    fi

    return $exit_val
}


#
# Export the global variable op_code
#
typeset -x op_code="${1}"
typeset AWK=/bin/awk

get_user_name_and_password_from_login_credentials_file
if [ $? -ne 0 ]; then
    sg_log 0 "ERROR: MSSQL Credential validation failed on the"\
             "node \"$LOCAL_HOST_NAME\""
    exit 1
fi

case "${op_code}" in
    "start")
        if sg_check_if_reconfig_has_failed ; then
            sg_log 999 "WARNING: Not attempting to start Microsoft SQL Server Always ON"\
                       "Availability instance on Linux"
            return
        else
            sg_log 0 "Starting Microsoft SQL Server Always ON Availability instance on Linux...."
            mssql_start_cmd $MSSQL_AVAILABILITY_GROUP
            check_status=$?
            if [[ $check_status -ne 0 ]] ; then
                if sg_check_if_it_is_reconfiguration ; then
                    sg_log 999 "ERROR: Failed to start Microsoft SQL Server Always ON Availability"\
                               "instance on Linux"
                    sg_set_reconfig_failure
                    exit_val=1
                else
                    to_exit=1
                    exit $to_exit
                fi
            else
               if sg_check_if_it_is_reconfiguration ; then
                   sg_log 999 "SUCCESS: Successfully started Microsoft SQL Server Always ON"\
                              "Availability instance on Linux"
               fi
            fi
        fi
        ;;
    "stop")
        sg_log 0 "Stopping Microsoft SQL Server Always ON Availability instance on Linux."
        mssql_stop_cmd
        ;;
    "validate")
        sg_log 0 "Validate the Microsoft SQL Availability instance AG parameters"
        mssql_validate_cmd
        ;;
    "reconfigure_start")
        reconfigure_start_function
        ;;
    "reconfigure_stop")
        reconfigure_stop_function
        ;;
    "replica_monitor")
        sg_log 0 "Starting Availability Local Instance Replica Monitoring for"\
                 "Microsoft SQL Server Always ON Availability instance on Linux."
        . $SGSBIN/mssql/instance_role.mon $*
        ;;
    *)
        sg_log 0 "Usage: ${0} [ start | stop | validate | reconfigure_start | reconfigure_stop | \
                                replica_monitor <replica_monitor_gr> ]"
        ;;
esac

exit $exit_val
